/*
 * @(#)cdemosecure.cpp    1.00   02/21/02
 *
 * Copyright (c) 2002-Present Agilis Software LLC, Santa Clara, CA, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Agilis Software LLC ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered
 * into with Agilis Software LLC.
 *
 * DISCLAIMER: this sample code is provided as-is.
 */

/*
 * This simple application illustrates the use of the secure license
 * key checking API of EasyLicenser and the digital signature
 * mechanism for protecting the application from spoofing.
 *
 * Security from spoofing is provided at the following levels:
 * 1. Using the custom key handler mechanism to implement a shared secret
 *    between the application, the EasyLicenser run time library and
 *    the license key.  The ISV embeds a custom key at the time of generating
 *    a license key.  The custom key typically contains information that is
 *    useful for license management purposes anyway - however, at its
 *    simplest level, it contains a "signature" indicating it's a key for
 *    the targeted application or application environment.  At run time,
 *    the ISV's application passes a custom key handler accompanied by
 *    a run time application context.  The application context may contain
 *    the application's signature.  The EasyLicenser invokes the custom
 *    key handler as part of its license key check, passing on the custom
 *    key embedded in the license key.  The application's implementation of
 *    the custom key handler can match the application context with the
 *    custom key to verify the identity of both, and post an event to that
 *    effect, whose result can be verified by the mainline code.
 *    If either the application code that invokes the EasyLicenser API is
 *    tampered with, or the EasyLicenser library is replaced with an impostor,
 *    the license key check fails.
 * 2. Using the EasyLicenser checkLicenseKeySecure API call to implement a
 *    dynamic handshake protocol.  At product installation time or the 
 *    first time the application is run, it obtains a seed key token from 
 *    the EasyLicenser run time library, and saves it in a secure persistent 
 *    location.  Every time a license key check is to be performed, for 
 *    example upon program startup, the saved key token is provided to 
 *    checkLicenseKeySecure, which validates it and produces and returns an 
 *    updated key token if the license key check is successful.  The 
 *    application replaces the previous key token with the new token.
 *    If the application code that invokes the EasyLicenser API is
 *    tampered with, or the system clock is turned back, or the
 *    key cookie is tampered with, the EasyLicenser 
 *    run time will fail the license key check regardless of the 
 *    validity of the license key.
 * 3. Digital signatures.  You can generate at your premises 
 *    your own password-protected digital signature of the 
 *    EasyLicenser library, or use the signature provided by Agilis.  
 *    Then embed or read in the signature and the public key encrypted
 *    password, and at run time recompute the signature based on the
 *    actual library and match it against the expected signature in 
 *    a spoof-proof manner using the public key encrypted forms of the 
 *    signature and password.  A library that has any alterations
 *    made to it will fail the check.
 *
 * The additional "application password" security mechanism is also
 * illustrated in C_Demo.c and cdemostate.c.
 *
 * The sample application illustrates the use of the above techniques.
 * In this simple example, the custom key handler doesn't enforce any 
 * event-driven protocol.  In a truly paranoid application, the custom key 
 * handler would post an event in the event of success.  The main program 
 * would define a listener on the event, and if after the license key check, 
 * the event had not been fired, that would imply a compromise in the 
 * security of one of the 3 participants.
 *
 * Summary of features illustrated with this demo:
 * 1. How to use the "checkLicenseKeySecure" API call and the key cookie
 *    mechanism to provide protection against system file changes.
 * 2. How to utilize the application-password security mechanism.
 * 3. How to utilize the digital signature mechanism.
 * 4. How to write and use a custom key handler.
 *
 * For examples of using the secure API for managing application
 * license state, see the accompanying "cdemostate.c" demo application.
 *
 * Some important points to note about this demo:
 * 1. The code that is directly relevant to using the EasyLicenser API's
 *    is indicated as such in comments.  The rest of the code exists in
 *    order to have a functioning self-contained program.
 * 2. All code pertaining to reading the key from an EasyLicenser file is
 *    for functional purposes only.  It is not meant to suggest a specific
 *    technique for managing the key in your application.  Nor is it
 *    a part of the EasyLicenser product.  In particular, note that your
 *    actual application will typically have access to just the license key
 *    without any accompanying metadata about the key.
 * 3. The demo will work with keys generated with EasyLicenser 2.0 as well
 *    as prior versions.  The EasyLicenser 2.0 run time library is required.
 * 4. This demo performs automatic state management in memory-resident 
 *    key cookies, rather than in a file that persists past its execution.
 *    Therefore, each successive run with the same parameters will have an 
 *    identical outcome with respect to quota-limits.
 * 5. Unlike the basic "C_Demo.c", this demo processes only the first key
 *    in its input file.  Therefore, provide a file containing only one key.
 *
 */

#define __STRICT_ANSI__ 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

// EasyLicenser header file
#include <ezLicenseInfo.h>
// EasyLicenser-included signature header file
#include <Signature.h>
// EasyLicenser-included encryption header file
#include <PkCrypt.h>


/* File types */
#define EZLM_FILE_TYPE_UNKNOWN	0
#define EZLM_FILE_TYPE_TXT		1
#define EZLM_FILE_TYPE_XML		2
#define EZLM_FILE_TYPE_CSV		3
#define EZLM_FILE_TYPE_TAB		4
#define EZLM_FILE_TYPE_INI		5
#define EZLM_FILE_TYPE_KEY		6


boolean CALLING_CONVENTION checkLibrarySignature(const char *sigFilename);
int CALLING_CONVENTION checkKeySecure(
							const char *keybuf,
							const char	*userhost,
							char		keycookie[],
							const char *productName,
							const char *appPwd,
							char		newkeycookie[],
							size_t		keycookiebufsize);

void CALLING_CONVENTION displayWarnings(int iWarningsBitMap);
void CALLING_CONVENTION displayError(int errorCode);
void CALLING_CONVENTION displayLicenseParams(void* ezLicInstance);
int CALLING_CONVENTION getFileType(const char *fname);
int CALLING_CONVENTION getKey(char *filelines[], int numlines,
							  int currLine, const int fileType,
							  char keybuf[], size_t keybufsize);
int CALLING_CONVENTION readFile(const char *filename, char *filelines[], int maxfilelines);
int CALLING_CONVENTION strip(const char *instr, char retbuf[], size_t retbufsize);
int customKeyHandler(void *keyCtx, 
					const char *strCustomKey);

/**
 * main entry point:
 * Accepts three mandatory command line arguments:
 * 1. Name of a text file containing a license key.
 * 2. Expected user / host name that should match what's in the key.
 *    If the key is known to have been generated with the "enforce OS
 *    user / host" flag, you can specify the special string "null" to
 *    signify a null value.
 * 3. The name of a file containing the EasyLicenser run time library's
 *    signature and accompanying password public key 
 *    based on a logical library name of "EzlmCppLibrary" and created
 *    with the platform-specific "makelibdigest" utility in 
 *    "encrypted, quiet" mode.  "makelibdigest" is located at
 *    the platform subdirectory of ezlic2.0/cpp/util/bin.  
 *    Alternatively, if you don't want to make your own signature,
 *    you can use the Agilis-supplied signature file "ezLicenserlib20-sig.txt" 
 *    also based on the same logical library name, and located at
 *    the platform-specific subdirectory of ezlic2.0/cpp/lib (the
 *    same directory containing the EasyLicenser run time library binary).
 * and two optional arguments (either none or both must be specified):
 * 4. Case-sensitive product name that was specified at the time of key creation,
 *    if the product had an associated application password
 * 5. Product's associated password's case-sensitive public key, available with
 *    the product definition and also obtainable with the "makepubkey" Java utility
 *    located at ezlic2.0/util.
 */
int main(int argc, char **argv)
{
	FILE* fptr = 0;
	char 	*strLicenseKeyFileName = NULL;
	char 	*strLicenseKey = NULL;
	int		iDaysLeftThreshold = 12;
	long	lQuotaUsageToDate = 0;
	long	lQuotaThreshold = 10;
	int		iDaysGracePeriod = 365; 
	long	lQuotaGrace = 100;
	long	lCurrUsageValue = 0;
	int		iStatusCode = 0;
	char	*strUserHostName;
	void *ezLicInfoInstance = 0;
	char *filelines[200];
	int numlines = 0;
	char keybuf[MAX_ENCRYPTED_LICENSE_TEXT_LENGTH];
	char keycookie1[MAX_KEY_COOKIE_SIZE];
	char keycookie2[MAX_KEY_COOKIE_SIZE];
	char *appPwd = NULL;
	char *productName = NULL;
	int fileType = EZLM_FILE_TYPE_UNKNOWN;
	int currLine = 0;
	int keyno = 0;
	int ret = 0;

    char    ezlmMajorVersion[MAX_VERSION_INFO_LENGTH];
    char    ezlmMinorVersion[MAX_VERSION_INFO_LENGTH];

	if (argc < 4 || argc == 5) {
		printf("\n*** usage: %s key-text-file-name user-or-host-name signature-file-name [product-name application-password-public-key] ***\n", argv[0]);
		printf("\n    Example: %s keypub.csv user1 ../../../lib/linux/ezLicenserlib20-sig.txt \n", argv[0]);
		exit(2);
	}

	strLicenseKeyFileName = argv[1];
	strUserHostName = argv[2];
	char *strSigfile = argv[3];
	if (argc > 4) {
		productName = argv[4];
		appPwd = argv[5];
	}

	printf("\nSecure License API Demo.\n");
    printf("\nRuntime library version information:\n");
    if (getEzlmVersionInfoStr(ezlmMajorVersion, sizeof(ezlmMajorVersion),
                              ezlmMinorVersion, sizeof(ezlmMinorVersion)) < 0) {
        printf("*** Error getting EasyLicenser runtime version information ***\n");
        exit (2);
    }
    printf("  Major Version: %s, Minor Version: %s\n", ezlmMajorVersion, ezlmMinorVersion);

	printf("\nCommand line arguments:\n");
	printf("  License key file:          %s\n", strLicenseKeyFileName);
	printf("  User / Host name:          %s\n", strUserHostName);
	printf("  Signature file:            %s\n", strSigfile);
	printf("  Product name:              %s\n", (productName == NULL ? "<not specified>":productName));
	printf("  App Passwd Public Key:     %s\n", (appPwd == NULL ? "<not specified>":appPwd));

	if (!checkLibrarySignature(strSigfile))
		exit(1);

	if (strcmp(strUserHostName,"null")==0)
		strUserHostName = NULL;

	fileType = getFileType(strLicenseKeyFileName);

	numlines = 
		readFile(strLicenseKeyFileName, filelines, 1000);
	if (numlines < 0) {
		printf("*** Cannot process license key file due to errors.  Aborting. ***\n");
		exit(1);
	} else if (!numlines) {
		printf("License key file is empty.  Nothing to do.  Exiting.\n");
		exit(0);
	}


	printf("\nGetting first license key in file...\n");
	strLicenseKey = &keybuf[0];
	if (getKey(filelines, numlines, currLine, fileType, keybuf, sizeof(keybuf)) <= 0)
		exit(1);

	printf("\nLicense key:\n");
	printf("[%s]\n\n", keybuf);

	printf("\nTry secure license key check sequence.\n");
	printf("1: Seed a key cookie.\n");
	// Simulates install time.
	// EasyLicenser API call to create an initial key cookie.
	if ((ret = createKeyCookieSeed_c(
						productName, 	/* product name for key */
						appPwd, 		/* application password public key */
						NULL, 			/* application state: none */
						TRUE,			/* verify system clock */
						keycookie1, 		/* buffer to receive key cookie */
						sizeof(keycookie1)) /* buffer size */
		) < 0) {
		printf("*** Error creating key cookie seed ***\n");
		displayError(ret);
		exit(2);
	}
	// after this, the installer saves this cookie safely, for example in 
	// a registry.

	printf("   Key cookie seed: [%s]\n", keycookie1);
	printf("2: Do secure key check with this cookie.  Should succeed if the key is good.\n");
	if (checkKeySecure(keybuf, strUserHostName, 
					   keycookie1, productName,
					   appPwd, keycookie2, 
					   sizeof(keycookie2)) < 0)
		goto failret;
	printf("3: Do secure key check with returned cookie.  Should succeed.\n");
	if (checkKeySecure(keybuf, strUserHostName, 
					   keycookie2, productName,
					   appPwd, keycookie1, 
					   sizeof(keycookie1)) < 0)
		goto failret;
	printf("4: Do secure key check with bad cookie.  Should fail.\n");
	strcpy(keycookie1, "oiuyaioufois");
	if (checkKeySecure(keybuf, strUserHostName, 
					   keycookie1, productName,
					   appPwd, keycookie2, 
					   sizeof(keycookie2)) >= 0) {
		printf("*** Failed: key check succeeded when it should have failed. ***\n");
		goto failret;
	} else
		printf("Ok.  Failed as expected.\n");
	return 0;
failret:
	printf("*** Demo failed ***\n");
	return 1;
}


int CALLING_CONVENTION checkKeySecure(
							const char *keybuf,
							const char	*userhost,
							char		keycookie[],
							const char *productName,
							const char *appPwd,
							char		newkeycookie[],
							size_t		keycookiebufsize)
{
	int callbackSignal = 0;
	// EasyLicenser API call: allocate a license context
	void *ezLicInfoInstance = allocLicenseInfo();
	// EasyLicenser API call: do secure license key check - full version
	int iStatusCode = checkLicenseKeySecure_stx(
								   ezLicInfoInstance,// license context
								   keybuf, 			 // license key string
								   keycookie,		 // the last-obtained key cookie
								   newkeycookie,	 // buffer to receive updated
													 // key cookie
								   keycookiebufsize, // output buffer size
								   customKeyHandler, // custom key handler
								   (void *)&callbackSignal,   // context for custom key handler
								   0, 				 // no expiration warning threshold
								   0,				 // no quota warning threshold
								   0, 				 // no quota increment
								   1000, 			 // allow license key check to
													 // succeed with warning instead of
													 // error within these many uses
													 // past quota limit.
								   0,			     // no quota grace - N/A w/ zero
													 // quota consumption
								   userhost, 		 // user / host name - can be null
													 // if OS enforcement is in effect.
								   0,				 // current usage level N/A.
								   productName,		 // name of product for which key
													 // was generated.
								   appPwd,			 // application password public key,
													 // if the product had an application
													 // password specified when the key was
													 // generated.
								   NULL,			 // no application state to save in
													 // key cookie.
                                   "ezlmhfile.ehf",  // explicit hidden file specification
                                   TRUE,             // check clock
								   TRUE);

	printf("\n returned from check lic with status code %d\n", iStatusCode);
	if (iStatusCode < 0)
	{
		displayError(iStatusCode);
	}
	else
	{	
		displayWarnings(iStatusCode);
		printf("\nSecure license key validation successful\n");
	}
	if (!callbackSignal) {
		printf("*** Library did not invoke callback ***\n");
	}
	// EasyLicenser API call: deallocate license context
	freeLicenseInfo(ezLicInfoInstance);
	return callbackSignal ? iStatusCode : EZLIC_EXC_AUTHENT;
}

boolean CALLING_CONVENTION checkLibrarySignature(const char *sigFilename) {
	char realEncrSig[SIGBUFSIZE];
	char encrPassword[SIGBUFSIZE];
	printf("Verifying signature of EasyLicenser library based on\ninfo in signature file '%s':\n", sigFilename);
	FILE *sf = fopen(sigFilename, "r");
	if (!sf) {
		printf("*** Signature file not found ***\n");
		return FALSE;
	}
	if (!fgets(realEncrSig, sizeof(realEncrSig), sf)) {
		printf("*** Invalid signature file ***\n");
		return FALSE;
	}
	if (isspace(realEncrSig[strlen(realEncrSig)-1]))
		realEncrSig[strlen(realEncrSig)-1] = '\0';
	if (!fgets(encrPassword, sizeof(encrPassword), sf)) {
		printf("*** Invalid signature file ***\n");
		return FALSE;
	}
	if (isspace(encrPassword[strlen(encrPassword)-1]))
		encrPassword[strlen(encrPassword)-1] = '\0';
	fclose(sf);
//	printf("Read signature file.\nSignature: [%s]\nEncr Signature: [%s]\nEncr Pwd: [%s]\n",
//		   sig, realEncrSig, encrPassword);
	char sigbuf[SIGBUFSIZE];
	char encsigbuf[SIGBUFSIZE];
	char encpwdbuf[SIGBUFSIZE];
	// EasyLicenser-included Signature API call: compute signature of
	// EasyLicenser run time library found at end-customer environment.
	if (makeLibDigest("ezLicenserlib20", encrPassword, 
					  "EzlmCppLibrary", FALSE,
					  sigbuf, sizeof(sigbuf),
					  encsigbuf, sizeof(encsigbuf),
					  encpwdbuf, sizeof(encpwdbuf))) {
		printf("Error getting library signature.  Details:\n");
		printf("%s\n", sigbuf);
		return FALSE;
	}
//	printf("Made lib digest.\nSignature: [%s]\nEncr Signature: [%s]\nEncr Pwd: [%s]\n",
//		   sigbuf, encsigbuf, encpwdbuf);
	char prodPubKey[SIGBUFSIZE];
	// EasyLicenser-included public key cryptography API call:
	// generate a public key using the product name as the seed.
	if (makePublicKey("EzlmCppLibrary", prodPubKey, sizeof(prodPubKey))) {
		printf("*** Error: could not make product public key ***\n");
		return FALSE;
	}
	char encrSig[SIGBUFSIZE];
	// EasyLicenser-included public key cryptography API call:
	// Encrypt the (already-encrypted) generated signature with the product's
	// public key.
	if (encryptWithPublicKey(sigbuf, prodPubKey, encrSig, sizeof(encrSig))) {
		printf("*** Error: could not encrypt signature \n[%s] with product public key '%s' ***\n", 
					   encrSig, prodPubKey);
		return FALSE;
	}
	if (strcmp(encrSig, encsigbuf)) {
		printf("*** Encrypted source signature doesn't match source encrypted signature - spoofed encryption library ***\n");
		return FALSE;
	}
	if (strcmp(encrSig, realEncrSig)) {
		printf("*** Encrypted source signature doesn't match input encrypted signature - spoofed product library ***\n");
//		printf("encrSig:\n[%s]\nrealEncrSig:\n[%s]\n", encrSig, realEncrSig);
		return FALSE;
	}
	printf("Signature check successful.\n");
	return TRUE;
}

/*
*	Display warnings if any.  Warning codes are defined in <ezLicStatusCodes.h>
*	@param iWarningsBitMap : Warning bitMap returned from license key API check.
*/
void CALLING_CONVENTION displayWarnings(int iWarningsBitMap)
{
	if (!iWarningsBitMap)
		return;
	printf("        Warning codes:\n");
    if (iWarningsBitMap & EZLIC_WARN_EXPIRING)
        printf("          Time limited license key is about to expire.\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA)
        printf("          Quota limited license key is about to expire.\n");
	if (iWarningsBitMap & EZLIC_WARN_EXPIRED)
        printf("          Time limited license has expired and is now in grace period.\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA_EXCEEDED)
        printf("          Quota limited license key has expired and is now in grace period.\n");
	if (iWarningsBitMap & EZLIC_WARN_EXP_GRACE_TRUNC)
		printf("          Expiration Days Grace In API Call Too Large - Truncated To 5 / 25 pct Of Lifespan\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA_GRACE_TRUNC)
		printf("          Quota Grace In API Call Too Large - Truncated To 5 / 25 pct Of Original Quota\n");
}


/*
*	Display the meaning of the -ve error code returned from the
*   license key API check.  Error codes are defined in <ezLicStatusCodes.h>
*/
void CALLING_CONVENTION displayError(int errorCode)
{
	if (!errorCode)
		return;

	printf("*** ");
	switch(errorCode) {
	case     EZLIC_EXC_BASE:
			printf("Error: General license validation failure. Please check your license key.\n");
			break;
	case     EZLIC_EXC_MALFORMED_KEY:
			printf("Error: the license key is invalid.\n");
			break;
	case     EZLIC_EXC_EXPIRED:
			printf("Error: The license has expired past the grace period.\n");
			break;
	case     EZLIC_EXC_EXCEEDED_LIMITS:
			printf("Error: The licensed user / server / CPU License / quota limit in the key has been exceeded.\n");
			break;
	case     EZLIC_EXC_INVALID_HOST_USER_NET:
			printf("Error: The user / machine name in the license key does not match your environment / parameter.\n");
			break;
	case     EZLIC_EXC_INVALID_CUSTOM_KEY:
			printf("Error: The custom key contained in the license key is invalid.\n");
			break;
	case     EZLIC_EXC_UNIMPLEMENTED:
			printf("Error: The feature you are using is not implemented.\n");
			break;
	case     EZLIC_EXC_GEN:
			printf("Error: Key generation execption (you shouldn't encounter this condition).\n");
			break;
	case     EZLIC_EXC_VENDOR:
			printf("Error: Vendor license key exception (you shouldn't encounter this condition).\n");
			break;
	case     EZLIC_EXC_INTERNAL:
			printf("Error: internal error in EasyLicenser - please submit a bug report.\n");
			break;
	case     EZLIC_EXC_OS:
			printf("Error: An operating system error occurred - please check your machine configuration.\n");
			break;
	case     EZLIC_EXC_CONFIG:
			printf("Error: exception in system configuration - please check your EasyLicenser installation and system clock.\n");
			break;
	case     EZLIC_EXC_UNLICENSED_OPTION:
			printf("Error: you attempted to perform an unlicensed operation.\n");
			break;
	case     EZLIC_EXC_USAGE:
			printf("Error: API usage - please check your API call parameters and call sequence.\n");
			break;
	case     EZLIC_EXC_AUTHENT:
			printf("Error: Key / cookie authentication failure - please check your product name and application password.\n");
			break;
	default:
			printf("Error: unknown error code %d.\n", errorCode);
			break;
	}
}

void CALLING_CONVENTION showLicenseModel(int licenseModel,
										 int licenseType)
{
	printf("        License Model: %s\n", 
			(licenseModel == EZLIC_MODEL_USER ? "User":
			 licenseModel == EZLIC_MODEL_SERVER ? "Server":
			 "*** Unknown ***"));
	printf("        License Type: ");
	switch(licenseModel) {
	case EZLIC_MODEL_USER:
		switch(licenseType) {
		case EZLIC_TYPE_U_USER:
			printf("Named User\n");
			break;
		case EZLIC_TYPE_U_NODE:
			printf("Node Locked\n");
			break;
		case EZLIC_TYPE_U_FLOAT:
			printf("Floating User\n");
			break;
		default:
			printf("*** Unknown license type %d ***\n", licenseType);
		}
		break;
	case EZLIC_MODEL_SERVER:
		switch(licenseType) {
		case EZLIC_TYPE_SVR_CONC:
			printf("Concurrent User\n");
			break;
		case EZLIC_TYPE_SVR_CPU:
			printf("CPU Count\n");
			break;
		case EZLIC_TYPE_SVR_MHZ:
			printf("CPU MHZ\n");
			break;
		case EZLIC_TYPE_SVR_NMU:
			printf("Named User\n");
			break;
		default:
			printf("*** Unknown license type %d ***\n", licenseType);
		}
		break;
	default:
		printf("%d\n", licenseModel);
	}
}


/*
* Introspect the license key and display its parameters.
* In an actual application you may introspect specific
* values and act on them instead of displaying them.
* The most common parameter that is post-processed is
* the product options.
*/
void CALLING_CONVENTION displayLicenseParams(
							void* ezLicInstance)
{
	int value = 0;
	long lvalue = 0L;
	long ltime = 0;
	char* strValue = 0;


	printf("    License Key Information:\n");
	printf("        Key version:    ");
	printf((getKeyVersion(ezLicInstance) <= 0 ? "Legacy (1.1 or earlier)\n":"%d\n"),
		   getKeyVersion(ezLicInstance));
	strValue = getChainedKeyId(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Replacement for key with Id: '%s'\n", strValue);

	value = getLicenseModelCode(ezLicInstance);
	showLicenseModel(value, getLicenseTypeCode(ezLicInstance));
	if (getEnforce(ezLicInstance))
		printf("        Operating System check enforced.\n");

	value = getLicenseModeBitmap(ezLicInstance);
	printf("        License Mode Bitmap:    0x%x\n", value);
	if (value & EZLIC_MODE_TIME)
		printf("          TIME LIMITED\n");
	if (value & EZLIC_MODE_OPTIONS)
		printf("          OPTIONS\n");
	if (value & EZLIC_MODE_METERED)
		printf("          METERED\n");
	if (value & EZLIC_MODE_CUSTOM_KEY)
		printf("          CUSTOM KEY\n");
	if (value & EZLIC_MODE_CUSTOM_COOKIE)
		printf("          CUSTOM COOKIE\n");

	lvalue = getUsageValue(ezLicInstance);
	if (lvalue > 0)
		printf("        Usage Limit Value:    %ld\n", lvalue);

	lvalue = getQuotaValue(ezLicInstance);
	if (lvalue > 0)
		printf("        Quota Limited To:    %ld\n", lvalue);

	strValue = getOptions(ezLicInstance);
	if (strValue  && strValue[0])
		printf("        Options:    [%s]\n", strValue);

	if ((ltime = getExpirationTime(ezLicInstance)) > 0) {
		char tmpbuf[100];
		struct tm when;
		time_t t = (time_t)ltime;
		when = *gmtime( &t );
		strcpy(tmpbuf,asctime(&when));
		if (tmpbuf[strlen(tmpbuf)-1]=='\n')
			tmpbuf[strlen(tmpbuf)-1] = '\0';
		printf( "        Time Limited To Expire (GMT):    %s\n", 
				tmpbuf);
	}
   	
	strValue = getUserHostNetName(ezLicInstance);
	if (strValue && strValue[0])
		printf("        User/Host Name:    [%s]\n", strValue);
	strValue = getCustomCookie(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Custom Cookie:    [%s]\n", strValue);
	strValue = getCustomKey(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Custom Key:    [%s]\n", strValue);
	strValue = getAppPwd(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Application Password:    [%s]\n", strValue);
}


/* 
 * The file IO is for convenience purposes only - you'd manage the persistence of 
 * your key according to the requirements of your application.
 */

int CALLING_CONVENTION readFile(const char *filename, char *filelines[], int maxfilelines)
{
	int numlines = 0;
	int totlen = 0;
	int len = 0;
	char linebuf[10240];
	char *linep = NULL;
	char *lp = NULL;
	int trunc = 0;
	FILE *fh = fopen(filename,"r");
	if (!fh) {
		printf("\n*** Cannot open license key file '%s' ***\n",
				filename);
		return -1;
	}
	while (numlines <= maxfilelines &&
	       fgets(linebuf, sizeof(linebuf), fh)) {
		linep = (char *)malloc(sizeof(linebuf));
		lp = &linebuf[0];
		while (*lp && isspace((int)*lp))
			lp++;
		if (*lp) {
			strcpy(linep, lp);
			filelines[numlines++] = linep;
		}
	}
	if ((!trunc) && numlines > maxfilelines)
		printf("\n*** Warning: IO buffer overflow, truncating.\n");
	fclose(fh);
	return numlines;
}


int CALLING_CONVENTION getKey(char *filelines[], int numlines,
							  int currLine, const int fileType,
							  char keybuf[], size_t keybufsize)
{
	const char *keyp = NULL;

	if (currLine >= numlines)
		return 0;

	if (currLine == 0) {
		// check / adjust for first line
		switch (fileType) {
		case EZLM_FILE_TYPE_TXT:
			if (strncmp(filelines[currLine++], "License Key List:", strlen("License Key List:")) ||
			    strncmp(filelines[currLine++], "----------------", strlen("----------------"))) {
				printf("*** Not a valid EasyLicenser text export file ***\n");
				exit(2);
			}
			currLine += 3; // position at "Key Id:"
			break;
		case EZLM_FILE_TYPE_XML:
			if (strncmp(filelines[currLine++], "<?xml", strlen("<?xml")) ||
			    strncmp(filelines[currLine++], "<key-list>", strlen("<key-list>"))) {
				printf("*** Not a valid EasyLicenser XML export file ***\n");
				exit(2);
			}
		default:
			break;
		}
	} else {
		// check for EOF
		switch (fileType) {
		case EZLM_FILE_TYPE_XML:
			if (filelines[currLine][0] && strncmp(filelines[currLine],"</key-list>",strlen("</key-list>"))==0)
				return 0;
		default:
			if (!filelines[currLine][0])
				return 0;
			break;
		}
	}
	// get key
	switch (fileType) {
	case EZLM_FILE_TYPE_TXT:
		for (; strncmp(filelines[currLine], "License key:", strlen("License key:")) && 
		       currLine < numlines; 
		     currLine++);
		if (++currLine >= numlines) {
			printf("*** Premature termination of text file ***\n");
			return 0;
		}
		keyp = filelines[currLine];
		strip(keyp, keybuf, keybufsize);
		currLine += 2;
		break;
	case EZLM_FILE_TYPE_XML:
		if (strncmp(filelines[currLine], "<key id=", strlen("<key id="))) {
			printf("*** Invalid XML record - treating as EOF ***\n");
			return 0;
		}
		if (++currLine >= numlines) {
			printf("*** Premature termination of XML file ***\n");
			return 0;
		}
		keyp = filelines[currLine];
		strip(keyp, keybuf, keybufsize);
		currLine++;
		currLine++;
		break;
	case EZLM_FILE_TYPE_INI:
		if (!(keyp = strchr(filelines[currLine], (int)'='))) {
			printf("*** Invalid property file format - treating as EOF ***\n");
			return 0;
		}
		strip(keyp+1, keybuf, keybufsize);
		currLine++;
		break;
	case EZLM_FILE_TYPE_KEY:
		strip(filelines[currLine], keybuf, keybufsize);
		currLine++;
		break;
	case EZLM_FILE_TYPE_CSV:
	case EZLM_FILE_TYPE_TAB:
		keyp = strrchr(filelines[currLine], (fileType==EZLM_FILE_TYPE_CSV?(int)',':(int)'\t'));
		if (!keyp) {
			printf("*** Invalid %s file format - treating as EOF ***\n",
					(fileType==EZLM_FILE_TYPE_CSV?"csv":"tab"));
			return 0;
		}
		strip(keyp+1, keybuf, keybufsize);
		currLine++;
		break;
	}
	return currLine;
}


int CALLING_CONVENTION getFileType(const char *fname) {
	char *suffix = strrchr((char *)fname, (int)'.');
	if (suffix > fname && suffix > &fname[strlen(fname)-4])
		return EZLM_FILE_TYPE_UNKNOWN;
	suffix++;
	if (strcmp(suffix, "txt")==0)
		return EZLM_FILE_TYPE_TXT;
	if (strcmp(suffix, "xml")==0)
		return EZLM_FILE_TYPE_XML;
	if (strcmp(suffix, "csv")==0)
		return EZLM_FILE_TYPE_CSV;
	if (strcmp(suffix, "tab")==0)
		return EZLM_FILE_TYPE_TAB;
	if (strcmp(suffix, "ini")==0)
		return EZLM_FILE_TYPE_INI;
	if (strcmp(suffix, "key")==0)
		return EZLM_FILE_TYPE_KEY;
	printf("*** Unknown file type: '%s'.  Exiting ***\n", suffix);
	exit(1);
}

/* 
 * Strip the key of all leading, trailing and embedded white spaces.
 * A good idea for an actual application, too.
 */
int CALLING_CONVENTION strip(const char *instr, char retbuf[], size_t retbufsize)
{
	char *outptr = &retbuf[0];
	int len = 0;
	char ch = 0;
	if (retbufsize <= strlen(instr))
		return -1;
	while (ch = *instr++) {
		if (!isspace((int)ch))
			*outptr++ = ch;
	}
	*outptr = '\0';
	return 0;
}

// Custom key handler: what's passed into the checkLicenseKey* API calls
// and invoked during license key checking.
int customKeyHandler(void *keyCtx, 
					const char *strCustomKey)
{
	int *signalp = (int *)keyCtx;
	if (strCustomKey && strCustomKey[0])
		printf("\n\t\tCustom key handler: custom key string is: %s\n",
				strCustomKey);
	printf("Setting signal.\n");
	*signalp = 1; // indicate we got here ok
	return 0;
}


