/*
 * @(#)jdemosecure.java    1.00   02/11/03
 *
 * Copyright (c) 2003++ Agilis Software LLC, Santa Clara, CA, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Agilis Software LLC ("Confidential Information").  You shall not
 * disclose such Confidential Information and shal use it only
 * in accordance with the terms of the license agreement you entered
 * into with Agilis Software LLC.
 *
 * DISCLAIMER: this sample code is provided as-is.
 */

/*
 * This simple application illustrates the use of the secure license
 * key checking API of EasyLicenser, digital signature mechanism and
 * dynamic invocation for protecting the application from spoofing.
 *
 * Security from spoofing is provided at the following levels:
 * 1. Using the custom key handler mechanism to implement a shared secret
 *    between the application, the EasyLicenser run time library and
 *    the license key.  The ISV embeds a custom key at the time of generating
 *    a license key.  The custom key typically contains information that is
 *    useful for license management purposes anyway - however, at its
 *    simplest level, it contains a "signature" indicating it's a key for
 *    the targeted application or application environment.  At run time,
 *    the ISV's application passes a custom key handler accompanied by
 *    a run time application context.  The application context may contain
 *    the application's signature.  The EasyLicenser invokes the custom
 *    key handler as part of its license key check, passing on the custom
 *    key embedded in the license key.  The application's implementation of
 *    the custom key handler can match the application context with the
 *    custom key to verify the identity of both, and post an event to that
 *    effect, whose result can be verified by the mainline code.
 *    If either the application code that invokes the EasyLicenser API is
 *    tampered with, or the EasyLicenser library is replaced with an impostor,
 *    the license key check fails.
 * 2. Using the EasyLicenser checkLicenseKeySecure API call to implement a
 *    dynamic handshake protocol.  At product installation time or the 
 *    first time the application is run, it obtains a seed key token from 
 *    the EasyLicenser run time library, and saves it in a secure persistent 
 *    location.  Every time a license key check is to be performed, for 
 *    example upon program startup, the saved key token is provided to 
 *    checkLicenseKeySecure, which validates it and produces and returns an 
 *    updated key token if the license key check is successful.  The 
 *    application replaces the previous key token with the new token.
 *    If the application code that invokes the EasyLicenser API is
 *    tampered with, or the system clock is turned back, or the
 *    key cookie is tampered with, the EasyLicenser 
 *    run time will fail the license key check regardless of the 
 *    validity of the license key.
 * 3. Digital signatures.  You can generate at your premises 
 *    your own password-protected digital signature of the 
 *    EasyLicenser library, or use the signature provided by Agilis.  
 *    Then embed or read in the signature and the public key encrypted
 *    password, and at run time recompute the signature based on the
 *    actual library and match it against the expected signature in 
 *    a spoof-proof manner using the public key encrypted forms of the 
 *    signature and password.  A library that has any alterations
 *    made to it will fail the check.
 *
 * The additional "application password" security mechanism is also
 * illustrated in jdemo.java and jdemostate.java.
 *
 * The sample application illustrates the use of the above techniques.
 * In this simple example, the custom key handler doesn't enforce any 
 * event-driven protocol.  In a truly paranoid application, the custom key 
 * handler would post an event in the event of success.  The main program 
 * would define a listener on the event, and if after the license key check, 
 * the event had not been fired, that would imply a compromise in the 
 * security of one of the 3 participants.
 *
 * Summary of features illustrated with this demo:
 * 1. How to use the "checkLicenseKeySecure" API call and the key cookie
 *    mechanism to provide protection against system file changes.
 * 2. How to utilize the application-password security mechanism.
 * 3. How to utilize the digital signature mechanism.
 * 4. How to write and use a custom key handler.
 * 5. How to use exception-based flow control and dynamic invocation
 *    to make it arbitrarily difficult to reverse engineer and alter 
 *    your program to bypass license key checks.
 *
 * For examples of using the secure API for managing application
 * license state, see the accompanying "jdemostate.java" demo application.
 *
 * Some important points to note about this demo:
 * 1. The code that is directly relevant to using the EasyLicenser API's
 *    is indicated as such in comments.  The rest of the code exists in
 *    order to have a functioning self-contained program.
 * 2. All code pertaining to reading the key from an EasyLicenser file is
 *    for functional purposes only.  It is not meant to suggest a specific
 *    technique for managing the key in your application.  Nor is it
 *    a part of the EasyLicenser product.  In particular, note that your
 *    actual application will typically have access to just the license key
 *    without any accompanying metadata about the key.
 * 3. The demo will work with keys generated with EasyLicenser 2.0 as well
 *    as prior versions.  The EasyLicenser 2.0 run time library is required.
 * 4. This demo performs automatic state management in memory-resident 
 *    key cookies, rather than in a file that persists past its execution.
 *    Therefore, each successive run with the same parameters will have an 
 *    identical outcome with respect to quota-limits.
 * 5. Unlike the basic "jdemo.java", this demo processes only the first key
 *    in its input file.  Therefore, provide a file containing only one key.
 *
 * More on exception based flow control:
 *
 *    The application defines an exception handler that inherits
 *    from an EasyLicenser base exception handler class, and passes an instance
 *    of it to the EasyLicenser checkLicenseKeySecure API call.
 *    When the license key check succeeds, the API call throws the custom
 *    exception handler instead of returning control to the caller.  When the
 *    exception handler is sufficiently distant from the API call in the 
 *    application code, it becomes harder for the application to be reverse
 *    engineered.
 *
 *    The sample application is necessarily small and does not attempt to
 *    distance the EasyLicenser API call from the success exception handler.
 *    In a real application, the exception handler would be far removed from
 *    the call.
 *
 * More on dynamic invocation:
 *
 *    The EasyLicenser license key checking API call is invoked through Java
 *    reflection instead of a straight method invocation.  The application
 *    uses variable names instead of literals so as to obfuscate the API
 *    call.  The dynamic invocation code is inlined with other code rather
 *    than isolated in order to be effective.  This is necessarily counter 
 *    to good programming practives.  
 *
 *    Dynamic invocation is a generic application programming technique that is 
 *    not unique to EasyLicenser.
 *
 */

import java.lang.reflect.*;
import java.io.*;
import java.text.*;
import java.util.*;
import com.vs.ils.crypt.*;
import com.vs.ezlicrun.*;

public class jdemosecure {

	static final int KEY_MAXLENGTH = 4096;

	// Key token.  In a real application, this would be maintained
	// in persistent store, for example a registry entry.
	static String dynClassname = "com.vs.ezlicrun.EzLicenseInfo";
	static String ezlmExceptionHandlerClassname = "com.vs.ezlicrun.EzLicExceptionBase";
	static String ezlmCustomKeyHandlerClassname = "com.vs.ezlicrun.EzLicCustomLicenseInterface";
	static String myExceptionHandlerClassname = "MyExceptionHandler1";
	static String myCustomKeyHandlerClassname = "MyCustomKeyHandler1";
	static String apiCallStr = "checkLicenseKeySecure";
	static String appsigStr = "jdemosecuresignature";

	/* File types */
	static final int EZLM_FILE_TYPE_UNKNOWN	= 0;
	static final int EZLM_FILE_TYPE_TXT		= 1;
	static final int EZLM_FILE_TYPE_XML		= 2;
	static final int EZLM_FILE_TYPE_CSV		= 3;
	static final int EZLM_FILE_TYPE_TAB		= 4;
	static final int EZLM_FILE_TYPE_INI		= 5;
	static final int EZLM_FILE_TYPE_KEY		= 6;

	public jdemosecure() {}

	/**
	 * main entry point:
 	 * Accepts three mandatory command line arguments:
 	 * 1. Name of a text file containing a license key.
 	 * 2. Expected user / host name that should match what's in the key.
 	 *    If the key is known to have been generated with the "enforce OS
 	 *    user / host" flag, you can specify the special string "null" to
 	 *    signify a null value.
 	 * 3. The name of a file containing the EasyLicenser run time library's
 	 *    signature and accompanying password public key 
 	 *    based on a logical library name of "EzlmCppLibrary" and created
 	 *    with the platform-specific "makelibdigest" utility in 
 	 *    "encrypted, quiet" mode.  "makelibdigest" is located at
 	 *    the platform subdirectory of ezlic2.0/cpp/util/bin.  
 	 *    Alternatively, if you don't want to make your own signature,
 	 *    you can use the Agilis-supplied signature file "ezLicenserlib25-sig.txt" 
 	 *    also based on the same logical library name, and located at
 	 *    the platform-specific subdirectory of ezlic2.0/cpp/lib (the
 	 *    same directory containing the EasyLicenser run time library binary).
 	 * and two optional arguments (either none or both must be specified):
 	 * 4. Case-sensitive product name that was specified at the time of key creation,
 	 *    if the product had an associated application password
 	 * 5. Product's associated password's case-sensitive public key, available with
 	 *    the product definition and also obtainable with the "makepubkey" Java utility
 	 *    located at ezlic2.0/util.
	 */
	public static void main(String[] args) {
		if (args.length < 3 || args.length == 4) {
			System.out.println("*** usage: jdemosecure key-text-file-name user-or-host-name signature-file-name [product-name application-password-public-key] ***");
			System.out.println("    Example: jdemosecure keypub.csv user1 ../../../lib/linux/ezLicenserlib25-sig.txt \n");
			System.exit(1);
		}
		String keyfile = args[0];
		String userhost = args[1];
		String sigFile = args[2];
		String productName = null;
		String appPwd = null;
		if (args.length > 3) {
			productName = args[3];
			appPwd = args[4];
		}

		System.out.println("\nSecure License API Demo.");
		System.out.println("Command line arguments:");
		System.out.println("  License key file:          " + keyfile);
		System.out.println("  User / Host name:          " + userhost);
		System.out.println("  Signature file:            " + sigFile);
		System.out.println("  Product name:              " + (productName == null ? "<not specified>":productName));
		System.out.println("  App Passwd Public Key:     " + (appPwd == null ? "<not specified>":appPwd));

		if (!checkLibrarySignature(sigFile))
			System.exit(1);

		if (userhost.equals("null"))
			userhost = null;
		int fileType = getFileType(keyfile);
		Object[] ret = readFile(keyfile, 1000);
		if (ret == null) {
			System.out.println("*** Cannot process license key file due to errors.  Aborting. ***");
			System.exit(1);
		}
		int numlines = ((Integer)ret[1]).intValue();
		if (numlines == 0) {
			System.out.println("License key file is empty.  Nothing to do.  Exiting.");
			System.exit(0);
		}

		System.out.println("Checking first license key in file...");
		String[] filelines = (String[])ret[0];
		Object[] ret2 = null;
		if ((ret2 = getKey(filelines, numlines, 0, fileType)) == null)
			System.exit(1);

		String keytext = (String)ret2[0];
		System.out.println("License key:\n["+keytext+"]\n");

		/* Simulate product and key installation: initialize key cookie. */
		String keyCookie = null;
		try {
			/* EasyLicenser API call: create an initial key cookie at product install time */
			keyCookie = EzLicenseInfo.createKeyCookieSeed(productName, appPwd, null, false);
		} catch (EzLicExceptionBase ex) {
			System.out.println("*** Error creating key cookie seed.  Exception: " + ex.getClass().getName() +
							   ", message: " + ex.getMessage());
			System.exit(2);
		}

		// First sequence of calls illustrates the use of the key cookie and
		// custom exception handling.

		// EasyLicenser API call (required): allocate a license context
		EzLicenseInfo licenseInfo = new EzLicenseInfo();
		int warnings = 0;
		try {
			// EasyLicenser API call (required): perform the license key check
			licenseInfo.checkLicenseKeySecure(
						keytext,		// the license key string
						keyCookie,		// previous key cookie
						(new MyExceptionHandler1()), // custom exception handler
						null, null,		// no custom key handler
						7,				// threshold days to trigger 
										// "about to expire" warning
						0, 				// no quota consumption
						10,				// threshold balance quota to
										// trigger "about to exhaust
										// quota" warning
						365, 			// allow license key check to
										// succeed with warning instead
										// of error within these many
										// days past expiration date.
						100,			// allow license key check to
										// succeed with warning instead
										// of error within these many
										// uses past quota limit.
						userhost, 		// user / host name - can be null
										// if OS enforcement is in effect.
						0,				// current use level (note: different
										// from quota which is cumulative
										// and can use different units) -
										// checked against limit in key.
										// applies only to non-single-user
										// licenses.
						productName,	// name of product for which key
										// was generated
						appPwd,			// application password public key,
										// if the product had an application
										// password specified when the key was
										// generated.
						null);			// no application state
		} catch (EzLicExceptionBase ex) {
			if (ex instanceof MyExceptionHandler1) {
				// in practice this would be caught somewhere in the code
				// quite removed from the API call.
				keyCookie = ex.getMessage(); 	// extract new key cookie
				warnings = ex.getMessageCode(); // extract warning codes
				checkWarnings(warnings);
				displayLicenseParams(licenseInfo);
			} else {
				System.out.println("*** License key check failed.  Error:\n\t" +
					ex.getMessage());
				System.exit(3);
			}
		}
		System.out.println("\nLicense key validation successful.\n");
		System.out.println("Re-checking the key to illustrate key cookie usage:");

		// Re-check the key, passing in the previously-returned key token,
		// to simulate a new key check at program startup.
		System.out.println("Re-checking the key to illustrate call w/o exception handler flow control.");
		System.out.println("This time, with a custom key handler.");
		warnings = 0;
		AppCtx appCtx = new AppCtx();
		try {
			// EasyLicenser API call (required): perform the license key check
			keyCookie =
			licenseInfo.checkLicenseKeySecure(
						keytext,		// the license key string
						keyCookie,		// previous key cookie
						null,			// no custom exception handler
						(new MyCustomKeyHandler1()), // custom key handler
						appCtx, 		// custom key context
						7,				// threshold days to trigger 
										// "about to expire" warning
						0, 				// no quota consumption
						10,				// threshold balance quota to
										// trigger "about to exhaust
										// quota" warning
						365, 			// allow license key check to
										// succeed with warning instead
										// of error within these many
										// days past expiration date.
						100,			// allow license key check to
										// succeed with warning instead
										// of error within these many
										// uses past quota limit.
						userhost, 		// user / host name - can be null
										// if OS enforcement is in effect.
						0,				// current use level (note: different
										// from quota which is cumulative
										// and can use different units) -
										// checked against limit in key.
										// applies only to non-single-user
										// licenses.
						productName,	// name of product for which key
										// was generated
						appPwd,			// application password public key,
										// if the product had an application
										// password specified when the key was
										// generated.
						null);			// no application state
		} catch (EzLicExceptionBase exxx) {
			System.out.println("*** Second license key check failed.  Error:\n    " +
				exxx.getMessage());
			System.exit(3);
		}

		if (!appCtx.signal) {
			System.out.println("*** Correct callback handler was not invoked ***");
		}

		System.out.println("Now do everything with dynamic invocation for further obfuscation.");
		appCtx.signal = false;
		try {
			Class dynClass = Class.forName(dynClassname);
			Class ezlmExceptionHandlerClass = Class.forName(ezlmExceptionHandlerClassname);
			Class ezlmCustomKeyHandlerClass = Class.forName(ezlmCustomKeyHandlerClassname);
			Class myExceptionHandlerClass = Class.forName(myExceptionHandlerClassname);
			Class myCustomKeyHandlerClass = Class.forName(myCustomKeyHandlerClassname);
			Object dyn = dynClass.newInstance();
			Object myexch = myExceptionHandlerClass.newInstance();
			Object mycsth = myCustomKeyHandlerClass.newInstance();
			Method methodObj = dynClass.getMethod(apiCallStr, 
										new Class[]
										{String.class,
										 String.class,
										 ezlmExceptionHandlerClass,
										 ezlmCustomKeyHandlerClass,
										 Object.class,
										 int.class, long.class, long.class,
										 int.class, long.class,
										 String.class, long.class,
										 String.class, String.class, String.class});
			Object[] callargs = { keytext, keyCookie, myexch, mycsth, 
							  appCtx, new Integer(7), new Long(0), 
							  new Long(10), new Integer(365), new Long(100),
							  userhost, new Long(0), productName, appPwd, null};
			methodObj.invoke(dyn, callargs);
		} catch (Exception ex) {
			Throwable t = null;
			if (ex instanceof InvocationTargetException)
				t = ((InvocationTargetException)ex).getTargetException();
			else
				t = ex;
			if (t instanceof MyExceptionHandler1) {
				System.out.println("OK.  Got success exception.");
			} else if (t instanceof Exception) {
				Exception exx = (Exception)t;
				System.out.println("*** Got unexpected exception.\n" +
							   "Exception class: " + exx.getClass().getName() +
							   "\nMessage: " + exx.getMessage());
				System.exit(1);
			} else {
				System.out.println("*** Caught unexpected throwable. ***");
				t.printStackTrace();
				System.exit(1);
			}
		}
		System.out.println("All license key validations successful.");
	}

	static void checkWarnings(int warnings)
	{
		if (warnings != 0) {
			System.out.println("    Warnings:");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_EXPIRING) != 0)
				System.out.println("        Time Limited Key Is About To Expire");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_QUOTA) != 0)
				System.out.println("        Quota Limited Key Is About To Expire");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_EXPIRED) != 0)
				System.out.println("        Time Limited Key Has Expired");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_QUOTA_EXCEEDED) != 0)
				System.out.println("        Quota Limited Key Has Expired");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_EXP_GRACE_TRUNC) != 0)
				System.out.println("        Expiration Days Grace In API Call Too Large - Truncated To 5 / 25% Of Lifespan");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_QUOTA_GRACE_TRUNC) != 0)
				System.out.println("        Quota Grace In API Call Too Large - Truncated To 5 / 25% Of Original Quota");
			System.out.println("");
		}
	}

	static void displayLicenseParams(EzLicenseInfo licenseInfo)
	{
		System.out.println("    License Key Information:");
		System.out.println("        Key Version: " + 
							(licenseInfo.getKeyVersion() <= 0 ? "Legacy (1.1 or earlier)" :
							 String.valueOf(licenseInfo.getKeyVersion()/10)+"." + 
							 String.valueOf(licenseInfo.getKeyVersion()%10)
							));
		if (licenseInfo.getChainedKeyId() != null)
			System.out.println("Replacement For Key With Id: " + licenseInfo.getChainedKeyId());
		showLicenseModel(licenseInfo);
		int bitmap = licenseInfo.getLicenseModeBitmap();
		System.out.println("        License Mode Bitmap: " + bitmap);
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_TIME) != 0)
			System.out.println("            TIME LIMITED");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_OPTIONS) != 0)
			System.out.println("            OPTIONS");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_METERED) != 0)
			System.out.println("            METERED");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_CUSTOM_KEY) != 0)
			System.out.println("            CUSTOM KEY");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_CUSTOM_COOKIE) != 0)
			System.out.println("            CUSTOM COOKIE");
		long lvalue;
		lvalue = licenseInfo.getUsageValue();
		if (lvalue > 0)
			System.out.println("        Usage Limit Value: " + lvalue);
		lvalue = licenseInfo.getQuotaValue();
		if (lvalue > 0)
			System.out.println("        Quota Value: " + lvalue);
		String options = licenseInfo.getOptions();
		if (options != null && !options.equals(""))
			System.out.println("        Options: [" + options + "]");
		Date expDate = licenseInfo.getExpirationDate();
		if (expDate != null)
			System.out.println("        Expiration Date: " +
							(new SimpleDateFormat("yyyy-MM-dd hh:mm:ss z")).
							format(expDate));
		System.out.println("        User / host name: " +
						licenseInfo.getUserHostNetName());
		if (licenseInfo.getEnforce())
			System.out.println("        Operating system check enforced.");
		String customCookie = licenseInfo.getCustomCookie();
		if (customCookie != null && !customCookie.equals(""))
			System.out.println("        Custom Cookie: [" + customCookie + "]");
		String customKey = licenseInfo.getCustomKey();
		if (customKey != null && !customKey.equals(""))
			System.out.println("        Custom Key: [" + customKey + "]");
		if (licenseInfo.getAppPwd() != null && !licenseInfo.getAppPwd().equals(""))
			System.out.println("        Application Password: [" + licenseInfo.getAppPwd() + "]");
	}

	static void showLicenseModel(EzLicenseInfo licenseInfo)
	{
		int licenseModel = licenseInfo.getLicenseModelCode();
		int licenseType = licenseInfo.getLicenseTypeCode();
		System.out.println("        License Model: " +
						   (licenseModel == EzLicenseInfo.EZLIC_MODEL_USER?"User":"Server"));
		System.out.print("        License Type: ");
		switch(licenseModel) {
		case EzLicenseInfo.EZLIC_MODEL_USER:
			switch(licenseType) {
			case EzLicenseInfo.EZLIC_TYPE_U_USER:
				System.out.println("Named User");
				break;
			case EzLicenseInfo.EZLIC_TYPE_U_NODE:
				System.out.println("Node Locked");
				break;
			case EzLicenseInfo.EZLIC_TYPE_U_FLOAT:
				System.out.println("Floating User");
				break;
			default:
				System.out.println("*** Unknown license type " + licenseType + " ***");
			}
			break;
		case EzLicenseInfo.EZLIC_MODEL_SERVER:
			switch(licenseType) {
			case EzLicenseInfo.EZLIC_TYPE_SVR_CONC:
				System.out.println("Concurrent User");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_CPU:
				System.out.println("CPU Count");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_MHZ:
				System.out.println("CPU MHZ");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_NMU:
				System.out.println("Named User");
				break;
			default:
				System.out.println("*** Unknown license type " + licenseType + " ***");
			}
			break;
		default:
			System.out.println("*** Unknown license model " + licenseModel + " ***");
		}
	}

	/* 
	 * The file IO is for convenience purposes only - you'd manage the persistence of 
	 * your key according to the requirements of your application.
	 */

	static Object[] readFile(String filename, int maxfilelines)
	{
		Object[] ret = new Object[2];
		String[] filelines = new String[maxfilelines];
		ret[0] = filelines;
		ret[1] = null;
		File kf = new File(filename);
		if (!kf.exists()) {
			System.out.println("*** License key file '" +
						filename + "' not found ***");
			System.exit(2);
		}
		int numlines = 0;
		BufferedReader in = null;
		try {
			in = new BufferedReader(
					new InputStreamReader(
						new FileInputStream(kf)));
			String keyline = null;
			while ((keyline = in.readLine()) != null && numlines < maxfilelines)
				filelines[numlines++] = keyline.trim();
			if (keyline != null) {
				System.out.println("*** Invalid key: too many lines.  Limit is " + maxfilelines + ". ***");
				System.exit(1);
			}
			filelines[numlines] = null;
		} catch (Exception ex) {
			System.out.println("*** IO Error reading license key file ***");
			System.out.println("Exception: " + ex.getClass().getName() + ", Message: " + ex.getMessage());
			//ex.printStackTrace();
			System.exit(2);
		} finally {
			try {
				in.close();
			} catch (Exception exx) {
			}
		}
		ret[1] = new Integer(numlines);
		return ret;
	}


	static Object[] getKey(String filelines[], int numlines,
						   int currLine, int fileType)
	{
		Object[] ret = new Object[2];
		ret[0] = null;
		ret[1] = null;
		int idx = -1;
		String keytext = null;

		if (currLine >= numlines)
			return null;

		if (currLine == 0) {
			/* check / adjust for first line */
			switch (fileType) {
			case EZLM_FILE_TYPE_TXT:
				if (!(filelines[currLine++].startsWith("License Key List:") &&
				      filelines[currLine++].startsWith("----------------"))) {
					System.out.println("*** Not a valid EasyLicenser text export file ***");
					System.out.println("Bad line #: " + currLine + ", contents: " + filelines[currLine-1]);
					System.exit(2);
				}
				currLine += 3; /* position at "Key Id:" */
				break;
			case EZLM_FILE_TYPE_XML:
				if (!(filelines[currLine++].startsWith("<?xml") &&
				      filelines[currLine++].startsWith("<key-list>"))) {
					System.out.println("*** Not a valid EasyLicenser XML export file ***");
					System.out.println("Bad line #: " + currLine + ", contents: " + filelines[currLine-1]);
					System.exit(2);
				}
			default:
				break;
			}
		} else {
			/* check for EOF */
			switch (fileType) {
			case EZLM_FILE_TYPE_XML:
				if (filelines[currLine] != null && 
				    filelines[currLine].startsWith("</key-list>"))
					return null;
			default:
				if (filelines[currLine] == null)
					return null;
				break;
			}
		}
		/* get key */
		switch (fileType) {
		case EZLM_FILE_TYPE_TXT:
			for (; !filelines[currLine].startsWith("License key:") && 
			       currLine < numlines; 
			     currLine++);
			if (++currLine >= numlines) {
				System.out.println("*** Premature termination of text file ***");
				return null;
			}
			keytext = filelines[currLine].trim();
			currLine += 2;
			break;
		case EZLM_FILE_TYPE_XML:
			if (!filelines[currLine].startsWith("<key id=")) {
				System.out.println("*** Invalid XML record - treating as EOF ***");
				System.out.println("Bad line #: " + currLine+1 + ", contents: " + filelines[currLine]);
				return null;
			}
			if (++currLine >= numlines) {
				System.out.println("*** Premature termination of XML file ***");
				return null;
			}
			keytext = filelines[currLine].trim();
			currLine++;
			currLine++;
			break;
		case EZLM_FILE_TYPE_INI:
			if ((idx = filelines[currLine].indexOf("=")) <= 0) {
				System.out.println("*** Invalid property file format - treating as EOF ***");
				return null;
			}
			keytext = filelines[currLine].substring(idx+1).trim();
			currLine++;
			break;
		case EZLM_FILE_TYPE_KEY:
			keytext = filelines[currLine].trim();
			currLine++;
			break;
		case EZLM_FILE_TYPE_CSV:
		case EZLM_FILE_TYPE_TAB:
			idx = filelines[currLine].lastIndexOf(
						(fileType==EZLM_FILE_TYPE_CSV?",":"\t"));
			if (idx <= 0) {
				System.out.println("*** Invalid " +
						(fileType==EZLM_FILE_TYPE_CSV?"csv":"tab") +
							" file format - treating as EOF ***");
				System.out.println("Bad line #: " + currLine+1 + ", contents: " + filelines[currLine]);
				return null;
			}
			keytext = filelines[currLine].substring(idx+1).trim();
			currLine++;
			break;
		}
		if (keytext.length() > KEY_MAXLENGTH) {
			System.out.println("*** Key length exceeds limit of " + KEY_MAXLENGTH + " characters ***");
			System.exit(2);
		}
		ret[0] = strip(keytext);
		ret[1] = new Integer(currLine);
		return ret;
	}

	static int getFileType(String filename) {
		int dotpos = filename.lastIndexOf('.');
		if (dotpos <= 0 || dotpos > filename.length()-4) {
			return EZLM_FILE_TYPE_UNKNOWN;
		}
		String suffix = filename.substring(dotpos+1);
		if (suffix.equals("txt"))
			return EZLM_FILE_TYPE_TXT;
		if (suffix.equals("xml"))
			return EZLM_FILE_TYPE_XML;
		if (suffix.equals("csv"))
			return EZLM_FILE_TYPE_CSV;
		if (suffix.equals("tab"))
			return EZLM_FILE_TYPE_TAB;
		if (suffix.equals("ini"))
			return EZLM_FILE_TYPE_INI;
		if (suffix.equals("key"))
			return EZLM_FILE_TYPE_KEY;
		System.out.println("*** Unknown file type: '" + suffix + "'.  Exiting ***");
		System.exit(1);
		return -1; // braindead compiler
	}

	static String strip(String instr) {
		String retstr = "";
		try {
			StringTokenizer stk = new StringTokenizer(instr," \n\t");
			while (stk.hasMoreTokens()) {
				String tokstr = stk.nextToken();
				retstr += tokstr;
			}
		} catch (Exception ex) {
			System.out.println("*** Got exception: " + ex.getMessage());
		}
		return retstr;
	}
	
	private static boolean checkLibrarySignature(String sigFilename) {
		System.out.println("Verifying signature of EasyLicenser run time library based on\n" +
						   "info in signature file '" + sigFilename + "':");
		File sf = new File(sigFilename);
		if (!sf.exists()) {
			System.out.println("*** Signature file not found ***");
			return false;
		}
		String realEncrSig = null;
		String encrPassword = null;
		BufferedReader in = null;
		try {
			in = new BufferedReader(new InputStreamReader(new FileInputStream(sf)));
			String linetxt = null;
			if ((linetxt = in.readLine()) == null) {
				System.out.println("*** Invalid signature file ***");
				return false;
			}
			realEncrSig = linetxt.trim();
			if ((linetxt = in.readLine()) == null) {
				System.out.println("*** Invalid signature file ***");
				return false;
			}
			encrPassword = linetxt.trim();
		} catch (Exception ex) {
			System.out.println("*** IO Error reading signature file ***");
			System.out.println("Exception: " + ex.getClass().getName() + ", Message: " + ex.getMessage());
			// ex.printStackTrace();
			return false;
		} finally {
			try {
				in.close();
			} catch (Exception exx) {
			}
		}
		String[] signatureEtc = 
			Signature.makeLibDigest("../lib/ezlicrun25.jar", encrPassword, 
									"EzlmJavaLibrary", false);
		if (signatureEtc[0].startsWith("Error")) {
			System.out.println("Error getting library signature.  Details:");
			System.out.println(signatureEtc[0]);
			return false;
		}
		String prodPubKey = com.vs.ils.crypt.PkCrypt.makePublicKey("EzlmJavaLibrary");
		String encrSig = com.vs.ils.crypt.PkCrypt.encryptWithPublicKey(signatureEtc[0], prodPubKey);
		if (!encrSig.equals(signatureEtc[1])) {
			System.out.println("*** Encrypted source signature doesn't match source encrypted signature - spoofed encryption library ***");
			return false;
		}
		if (!encrSig.equals(realEncrSig)) {
			System.out.println("*** Encrypted source signature doesn't match input encrypted signature - spoofed product library ***");
			return false;
		}
		System.out.println("Signature check successful.");
		return true;
	}

}
// Custom exception handler for signalling success
class MyExceptionHandler1 extends EzLicExceptionBase {
	MyExceptionHandler1() {}
}
// Custom key handler
class MyCustomKeyHandler1 implements EzLicCustomLicenseInterface {
	
	public int checkCustomKeyCode(String valStr, Object ctx) 
			throws EzLicExceptionBase
	{
		// checks for whether the run time context signature matches 
		// the expected application signature.  Should also match
		// it against the custom key content in a real application
		// for added security.
		AppCtx appCtx = (AppCtx)ctx;
		if (!appCtx.signature.equals(jdemosecure.appsigStr)) {
			System.out.println("*** Signature mismatch.  "+appCtx.signature+" vs. " + jdemosecure.appsigStr + " ***");
			System.out.println("Possible compromise in 3-way secret.");
			// should throw an exception here
		} else
			System.out.println("Signatures match and all is well.");
			// should also post a success event here
		appCtx.signal = true; // indicate we got here
		return 0;
	}

}

// Application context class to pass into custom key handler
class AppCtx {
	boolean		signal = false;
	String		signature = jdemosecure.appsigStr;
}
